@@ -25,6 +25,7 @@
 
 #include <stdio.h>
 #include <string.h>
+#include <stdlib.h>
 #include <gdbus.h>
 
 #include "connman.h"
@@ -35,6 +36,13 @@ static GHashTable *provider_hash = NULL;
 
 static GSList *driver_list = NULL;
 
+struct connman_route {
+	int family;
+	char *host;
+	char *netmask;
+	char *gateway;
+};
+
 struct connman_provider {
 	struct connman_element element;
 	struct connman_service *vpn_service;
@@ -44,6 +52,7 @@ struct connman_provider {
 	char *host;
 	char *dns;
 	char *domain;
+	GHashTable *routes;
 	struct connman_provider_driver *driver;
 	void *driver_data;
 };
@@ -84,6 +93,9 @@ static int connman_provider_setup_vpn_ipv4(struct connman_provider *provider,
 	g_free(element->ipv4.address);
 	element->ipv4.address = g_strdup(provider->element.ipv4.address);
 
+	g_free(element->ipv4.peer);
+	element->ipv4.peer = g_strdup(provider->element.ipv4.peer);
+
 	g_free(element->ipv4.netmask);
 	element->ipv4.netmask = g_strdup(provider->element.ipv4.netmask);
 
@@ -165,7 +177,8 @@ int __connman_provider_disconnect(struct connman_provider *provider)
 	else
 		return -EOPNOTSUPP;
 
-	__connman_service_indicate_state(provider->vpn_service,
+	if (provider->vpn_service != NULL)
+		__connman_service_indicate_state(provider->vpn_service,
 					CONNMAN_SERVICE_STATE_DISCONNECT);
 	if (err < 0) {
 		if (err != -EINPROGRESS)
@@ -184,12 +197,14 @@ int __connman_provider_connect(struct connman_provider *provider)
 	DBG("provider %p", provider);
 
 	g_free(provider->element.ipv4.address);
+	g_free(provider->element.ipv4.peer);
 	g_free(provider->element.ipv4.netmask);
 	g_free(provider->element.ipv4.gateway);
 	g_free(provider->element.ipv4.broadcast);
 	g_free(provider->element.ipv4.pac);
 
 	provider->element.ipv4.address = NULL;
+	provider->element.ipv4.peer = NULL;
 	provider->element.ipv4.netmask = NULL;
 	provider->element.ipv4.gateway = NULL;
 	provider->element.ipv4.broadcast = NULL;
@@ -215,18 +230,87 @@ int __connman_provider_connect(struct connman_provider *provider)
 int __connman_provider_remove(const char *path)
 {
 	struct connman_provider *provider;
+	GHashTableIter iter;
+	gpointer value, key;
 
 	DBG("path %s", path);
 
-	provider = g_hash_table_lookup(provider_hash, path);
-	if (provider == NULL) {
-		DBG("patch %s not found", path);
-		return -ENXIO;
+	g_hash_table_iter_init(&iter, provider_hash);
+	while (g_hash_table_iter_next(&iter, &key, &value) == TRUE) {
+		const char *srv_path;
+		provider = value;
+
+		if (provider->vpn_service == NULL)
+			continue;
+
+		srv_path = __connman_service_get_path(provider->vpn_service);
+
+		if (g_strcmp0(srv_path, path) == 0) {
+			DBG("Removing VPN %s", provider->identifier);
+			g_hash_table_remove(provider_hash,
+						provider->identifier);
+			return 0;
+		}
 	}
 
-	g_hash_table_remove(provider_hash, path);
+	return -ENXIO;
+}
 
-	return 0;
+static void provider_append_routes(gpointer key, gpointer value,
+					gpointer user_data)
+{
+	struct connman_route *route = value;
+	struct connman_provider *provider = user_data;
+	int index = provider->element.index;
+
+	if (route->family == AF_INET6) {
+		unsigned char prefix_len = atoi(route->netmask);
+
+		connman_inet_add_ipv6_network_route(index, route->host,
+							route->gateway,
+							prefix_len);
+	} else {
+		connman_inet_add_network_route(index, route->host,
+						route->gateway,
+						route->netmask);
+	}
+}
+
+static void provider_set_nameservers(struct connman_provider *provider)
+{
+	struct connman_service *service = provider->vpn_service;
+
+	char *nameservers = NULL, *name = NULL;
+	const char *value;
+	char *second_ns;
+
+	if (service == NULL)
+		return;
+
+	if (provider->dns == NULL)
+		return;
+
+	name = connman_inet_ifname(provider->element.index);
+
+	nameservers = g_strdup(provider->dns);
+	value = nameservers;
+	second_ns = strchr(value, ' ');
+	if (second_ns)
+		*(second_ns++) = 0;
+	__connman_service_append_nameserver(service, value);
+	value = second_ns;
+
+	while (value) {
+		char *next = strchr(value, ' ');
+		if (next)
+			*(next++) = 0;
+
+		connman_resolver_append(name, provider->domain, value);
+		value = next;
+	}
+
+	g_free(nameservers);
+	g_free(name);
 }
 
 static int set_connected(struct connman_provider *provider,
@@ -240,9 +324,6 @@ static int set_connected(struct connman_provider *provider,
 	if (connected == TRUE) {
 		enum connman_element_type type = CONNMAN_ELEMENT_TYPE_UNKNOWN;
 		struct connman_element *element;
-		char *nameservers = NULL, *name = NULL;
-		const char *value;
-		char *second_ns;
 		int err;
 
 		__connman_service_indicate_state(provider->vpn_service,
@@ -272,28 +353,10 @@ static int set_connected(struct connman_provider *provider,
 
 		__connman_service_set_domainname(service, provider->domain);
 
-		name = connman_inet_ifname(provider->element.index);
-
-		nameservers = g_strdup(provider->dns);
-		value = nameservers;
-		second_ns = strchr(value, ' ');
-		if (second_ns)
-			*(second_ns++) = 0;
-		__connman_service_append_nameserver(service, value);
-		value = second_ns;
-
-		while (value) {
-			char *next = strchr(value, ' ');
-			if (next)
-				*(next++) = 0;
-
-			connman_resolver_append(name, provider->domain, value);
-			value = next;
-		}
-
-		g_free(nameservers);
-		g_free(name);
+		provider_set_nameservers(provider);
 
+		g_hash_table_foreach(provider->routes, provider_append_routes,
+					provider);
 	} else {
 		connman_element_unregister_children(&provider->element);
 		__connman_service_indicate_state(service,
@@ -330,13 +393,39 @@ int connman_provider_set_state(struct connman_provider *provider,
 	return -EINVAL;
 }
 
+int connman_provider_indicate_error(struct connman_provider *provider,
+					enum connman_provider_error error)
+{
+	enum connman_service_error service_error;
+
+	switch (error) {
+	case CONNMAN_PROVIDER_ERROR_LOGIN_FAILED:
+		service_error = CONNMAN_SERVICE_ERROR_LOGIN_FAILED;
+		break;
+	case CONNMAN_PROVIDER_ERROR_AUTH_FAILED:
+		service_error = CONNMAN_SERVICE_ERROR_AUTH_FAILED;
+		break;
+	case CONNMAN_PROVIDER_ERROR_CONNECT_FAILED:
+		service_error = CONNMAN_SERVICE_ERROR_CONNECT_FAILED;
+		break;
+	default:
+		service_error = CONNMAN_SERVICE_ERROR_UNKNOWN;
+		break;
+	}
+
+	return __connman_service_indicate_error(provider->vpn_service,
+							service_error);
+}
+
 static void unregister_provider(gpointer data)
 {
 	struct connman_provider *provider = data;
+	struct connman_service *service = provider->vpn_service;
 
 	DBG("provider %p", provider);
 
-	__connman_service_put(provider->vpn_service);
+	provider->vpn_service = NULL;
+	__connman_service_put(service);
 
 	connman_element_unregister(&provider->element);
 	connman_provider_unref(provider);
@@ -353,6 +442,17 @@ static void provider_destruct(struct connman_element *element)
 	g_free(provider->domain);
 	g_free(provider->identifier);
 	g_free(provider->dns);
+	g_hash_table_destroy(provider->routes);
+}
+
+static void destroy_route(gpointer user_data)
+{
+	struct connman_route *route = user_data;
+
+	g_free(route->host);
+	g_free(route->netmask);
+	g_free(route->gateway);
+	g_free(route);
 }
 
 static void provider_initialize(struct connman_provider *provider)
@@ -375,6 +475,8 @@ static void provider_initialize(struct connman_provider *provider)
 	provider->dns = NULL;
 	provider->domain = NULL;
 	provider->identifier = NULL;
+	provider->routes = g_hash_table_new_full(g_direct_hash, g_direct_equal,
+					NULL, destroy_route);
 }
 
 static struct connman_provider *connman_provider_new(void)
@@ -537,7 +639,8 @@ int __connman_provider_create_and_connect(DBusMessage *msg)
 	if (created == TRUE)
 		provider_probe(provider);
 
-	provider->vpn_service =
+	if (provider->vpn_service == NULL)
+		provider->vpn_service =
 			__connman_service_create_from_provider(provider);
 	if (provider->vpn_service == NULL) {
 		err = -EOPNOTSUPP;
@@ -559,8 +662,10 @@ failed:
 		DBG("can not connect delete provider");
 		connman_provider_unref(provider);
 
-		if (provider->vpn_service != NULL)
+		if (provider->vpn_service != NULL) {
 			__connman_service_put(provider->vpn_service);
+			provider->vpn_service = NULL;
+		}
 	}
 
 	return err;
@@ -591,6 +696,9 @@ int connman_provider_set_string(struct connman_provider *provider,
 	} else if (g_str_equal(key, "Address") == TRUE) {
 		g_free(provider->element.ipv4.address);
 		provider->element.ipv4.address = g_strdup(value);
+	} else if (g_str_equal(key, "Peer") == TRUE) {
+		g_free(provider->element.ipv4.peer);
+		provider->element.ipv4.peer = g_strdup(value);
 	} else if (g_str_equal(key, "Netmask") == TRUE) {
 		g_free(provider->element.ipv4.netmask);
 		provider->element.ipv4.netmask = g_strdup(value);
@@ -640,12 +748,15 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 
 	DBG("");
 
-	ipconfig = __connman_service_get_ipconfig(service);
+	if (service == NULL)
+		return;
+
+	ipconfig = __connman_service_get_ip4config(service);
 
 	if (ipconfig == NULL) {
-		__connman_service_create_ipconfig(service, index);
+		__connman_service_create_ip4config(service, index);
 
-		ipconfig = __connman_service_get_ipconfig(service);
+		ipconfig = __connman_service_get_ip4config(service);
 		if (ipconfig == NULL) {
 			DBG("Couldnt create ipconfig");
 			goto done;
@@ -656,6 +767,21 @@ void connman_provider_set_index(struct connman_provider *provider, int index)
 	__connman_ipconfig_set_index(ipconfig, index);
 
 
+	ipconfig = __connman_service_get_ip6config(service);
+
+	if (ipconfig == NULL) {
+		__connman_service_create_ip6config(service, index);
+
+		ipconfig = __connman_service_get_ip6config(service);
+		if (ipconfig == NULL) {
+			DBG("Couldnt create ipconfig for IPv6");
+			goto done;
+		}
+	}
+
+	connman_ipconfig_set_method(ipconfig, CONNMAN_IPCONFIG_METHOD_OFF);
+	__connman_ipconfig_set_index(ipconfig, index);
+
 done:
 	provider->element.index = index;
 }
@@ -665,6 +791,116 @@ int connman_provider_get_index(struct connman_provider *provider)
 	return provider->element.index;
 }
 
+enum provider_route_type {
+	PROVIDER_ROUTE_TYPE_NONE = 0,
+	PROVIDER_ROUTE_TYPE_MASK = 1,
+	PROVIDER_ROUTE_TYPE_ADDR = 2,
+	PROVIDER_ROUTE_TYPE_GW   = 3,
+};
+
+static int route_env_parse(struct connman_provider *provider, const char *key,
+				int *family, unsigned long *idx,
+				enum provider_route_type *type)
+{
+	char *end;
+	const char *start;
+
+	DBG("name %s", provider->name);
+
+	if (!strcmp(provider->type, "openvpn")) {
+		if (g_str_has_prefix(key, "route_network_") == TRUE) {
+			start = key + strlen("route_network_");
+			*type = PROVIDER_ROUTE_TYPE_ADDR;
+		} else if (g_str_has_prefix(key, "route_netmask_") == TRUE) {
+			start = key + strlen("route_netmask_");
+			*type = PROVIDER_ROUTE_TYPE_MASK;
+		} else if (g_str_has_prefix(key, "route_gateway_") == TRUE) {
+			start = key + strlen("route_gateway_");
+			*type = PROVIDER_ROUTE_TYPE_GW;
+		} else
+			return -EINVAL;
+
+		*family = AF_INET;
+		*idx = g_ascii_strtoull(start, &end, 10);
+
+	} else if (!strcmp(provider->type, "openconnect")) {
+		if (g_str_has_prefix(key, "CISCO_SPLIT_INC_") == TRUE) {
+			*family = AF_INET;
+			start = key + strlen("CISCO_SPLIT_INC_");
+		} else if (g_str_has_prefix(key, "CISCO_IPV6_SPLIT_INC_") == TRUE) {
+			*family = AF_INET6;
+			start = key + strlen("CISCO_IPV6_SPLIT_INC_");
+		} else
+			return -EINVAL;
+
+		*idx = g_ascii_strtoull(start, &end, 10);
+
+		if (strncmp(end, "_ADDR", 5) == 0)
+			*type = PROVIDER_ROUTE_TYPE_ADDR;
+		else if (strncmp(end, "_MASK", 5) == 0)
+			*type = PROVIDER_ROUTE_TYPE_MASK;
+		else if (strncmp(end, "_MASKLEN", 8) == 0 &&
+				*family == AF_INET6) {
+			*type = PROVIDER_ROUTE_TYPE_MASK;
+		} else
+			return -EINVAL;
+	}
+
+	return 0;
+}
+
+int connman_provider_append_route(struct connman_provider *provider,
+					const char *key, const char *value)
+{
+	struct connman_route *route;
+	int ret, family = 0;
+	unsigned long idx = 0;
+	enum provider_route_type type = PROVIDER_ROUTE_TYPE_NONE;
+
+	DBG("key %s value %s", key, value);
+
+	ret = route_env_parse(provider, key, &family, &idx, &type);
+	if (ret < 0)
+		return ret;
+
+	DBG("idx %lu family %d type %d", idx, family, type);
+
+	route = g_hash_table_lookup(provider->routes, GINT_TO_POINTER(idx));
+	if (route == NULL) {
+		route = g_try_new0(struct connman_route, 1);
+		if (route == NULL) {
+			connman_error("out of memory");
+			return -ENOMEM;
+		}
+
+		route->family = family;
+
+		g_hash_table_replace(provider->routes, GINT_TO_POINTER(idx),
+						route);
+	}
+
+	switch (type) {
+	case PROVIDER_ROUTE_TYPE_NONE:
+		break;
+	case PROVIDER_ROUTE_TYPE_MASK:
+		route->netmask = g_strdup(value);
+		break;
+	case PROVIDER_ROUTE_TYPE_ADDR:
+		route->host = g_strdup(value);
+		break;
+	case PROVIDER_ROUTE_TYPE_GW:
+		route->gateway = g_strdup(value);
+		break;
+	}
+
+	return 0;
+}
+
+const char *connman_provider_get_driver_name(struct connman_provider *provider)
+{
+	return provider->driver->name;
+}
+
 static gint compare_priority(gconstpointer a, gconstpointer b)
 {
 	return 0;
@@ -694,12 +930,12 @@ void connman_provider_driver_unregister(struct connman_provider_driver *driver)
 	driver_list = g_slist_remove(driver_list, driver);
 }
 
-static void provider_disconnect(gpointer key, gpointer value,
+static void provider_remove(gpointer key, gpointer value,
 						gpointer user_data)
 {
 	struct connman_provider *provider = value;
 
-	__connman_provider_disconnect(provider);
+	g_hash_table_remove(provider_hash, provider->identifier);
 }
 
 static void provider_offline_mode(connman_bool_t enabled)
@@ -707,7 +943,7 @@ static void provider_offline_mode(connman_bool_t enabled)
 	DBG("enabled %d", enabled);
 
 	if (enabled == TRUE)
-		g_hash_table_foreach(provider_hash, provider_disconnect, NULL);
+		g_hash_table_foreach(provider_hash, provider_remove, NULL);
 
 }
 
